/*
 *  Copyright (C) Chaperon. All rights reserved.                               
 *  ------------------------------------------------------------------------- 
 *  This software is published under the terms of the Apache Software License 
 *  version 1.1, a copy of which has been included  with this distribution in 
 *  the LICENSE file.                                                         
 */

package net.sourceforge.chaperon.parser.generator;

import java.util.ArrayList;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.AttributesImpl;
import net.sourceforge.chaperon.grammar.Grammar;
import net.sourceforge.chaperon.grammar.symbol.Symbol;
import net.sourceforge.chaperon.grammar.symbol.SymbolList;

/**
 * This class contains a collection of item sets (states)
 *
 * @author Stephan Michels
 * @version CVS $Id: Collection.java,v 1.4 2002/04/11 14:53:55 benedikta Exp $
 */
public class Collection
{

  private ArrayList _states = new ArrayList();

  private Grammar   _grammar;

  /**
   * Create a collection of itemsets
   *
   * @param grammar Grammar
   */
  public Collection(Grammar grammar)
  {
    _grammar = grammar;
  }

  /**
   * Add a itemset to this collection
   *
   * @param itemset Itemset
   *
   * @return Index of the itemset in the collection
   */
  public int add(ItemSet itemset)
  {
    int index = indexOf(itemset);

    if (index == -1)
    {
      _states.add(itemset);
      index = _states.size() - 1;
    }
    return index;
  }

  /**
   * Return an itemset by an index
   *
   * @param index Index of the itemset
   *
   * @return Itemset
   */
  public ItemSet get(int index)
  {
    return (ItemSet) _states.get(index);
  }

  /**
   * Return the count of itemsets(states) in this collection
   *
   * @return Count of itemsets
   */
  public int getSize()
  {
    return _states.size();
  }

  /**
   * Return the index of an item set. If the collection does
   * not contain the itemset, then return this method -1
   *
   * @param itemset Itemset
   *
   * @return Index of the itemset
   */
  public int indexOf(ItemSet itemset)
  {
    for (int i = 0; i < _states.size(); i++)
      if (((ItemSet) _states.get(i)).equals(itemset))
        return i;
    return -1;
  }

  /**
   * If this collection contains the item set
   *
   * @param itemset Itemset
   *
   * @return True, if the collection conatins the itemset
   */
  public boolean contains(ItemSet itemset)
  {
    return (indexOf(itemset) >= 0);
  }

  /**
   * Return the index of an item set. If the collection
   * does not contain the itemset, then return this method -1.
   * This Method compare only the core from the itemset.
   *
   * @param itemset Itemset
   *
   * @return Index of the Itemset
   */
  public int indexOfCore(ItemSet itemset)
  {
    for (int i = 0; i < _states.size(); i++)
      if (((ItemSet) _states.get(i)).equalsCore(itemset))
        return i;
    return -1;
  }

  /**
   * If this collection contains the item set. This method
   * compares only the core of the itemset
   *
   * @param itemset Itemset
   *
   * @return True, if the collection contain the itemset
   */
  public boolean containsCore(ItemSet itemset)
  {
    return (indexOfCore(itemset) >= 0);
  }

  /**
   * Converts the collection to a string
   *
   * @return Content of the collection
   */
  public String toString()
  {
    StringBuffer buffer = new StringBuffer();

    for (int i = 0; i < getSize(); i++)
    {
      buffer.append("State " + i + ":\n");
      buffer.append(get(i).toString());
      buffer.append("\n");
    }
    return buffer.toString();
  }
}
